REDIS_URL_ENV = "REDIS_URL"
REDIS_HOST_ENV = "REDIS_HOST"
REDIS_PORT_ENV = "REDIS_PORT"
REDIS_DB_ENV = "REDIS_DB"
REDIS_CLUSTER_ENV = "REDIS_CLUSTER"
REDIS_TLS_ENV = "REDIS_TLS"
REDIS_USERNAME_ENV = "REDIS_USERNAME"
REDIS_PASSWORD_ENV = "REDIS_PASSWORD"


REDIS_PREFIX_ENV = "REDIS_PREFIX"
REDIS_SEPERATOR_ENV = "REDIS_SEPERATOR"
